from io import BytesIO
from fittie.fitfile.records import read_record_header


def test_record_header() -> None:
    # Definition message
    data = BytesIO(initial_bytes=0b01000000.to_bytes(length=1, byteorder="big"))
    header = read_record_header(data)
    assert header.is_definition_message
    assert header.local_message_type == 0
    assert not header.is_developer_data

    # Data message
    data = BytesIO(initial_bytes=0b00000000.to_bytes(length=1, byteorder="big"))
    header = read_record_header(data)
    assert not header.is_definition_message
    assert header.local_message_type == 0
    assert not header.is_developer_data

    # Definition message with developer data
    data = BytesIO(initial_bytes=0b01100001.to_bytes(length=1, byteorder="big"))
    header = read_record_header(data)
    assert header.is_definition_message
    assert header.local_message_type == 1
    assert header.is_developer_data

    # Compressed timestamp message
    data = BytesIO(initial_bytes=0b10000000.to_bytes(length=1, byteorder="big"))
    header = read_record_header(data)
    assert not header.is_definition_message
    assert not header.local_message_type == 1
    assert not header.is_developer_data
    assert header.is_compressed_timestamp_message



from __future__ import annotations

import struct
from typing import Optional

from fittie.utils.datastream import Streamable


class BaseType:
    """Representation of a Garmin FIT File Base Type"""

    number: int
    endian_ability: int
    name: str
    invalid_value: int
    size: int
    fmt: str
    value_type: type
    comment: str

    def __init__(
        self,
        number: int,
        endian_ability: int,
        name: str,
        invalid_value: int,
        size: int,
        fmt: str,
        value_type: type,
        comment: Optional[str] = None,
    ):
        self.number = number
        self.endian_ability = endian_ability
        self.name = name
        self.invalid_value = invalid_value
        self.size = size
        self.fmt = fmt
        self.value_type = value_type
        self.comment = comment if comment is not None else ""

    def __str__(self) -> str:
        return f"BaseType:{self.name=}{self.number=}".replace("self.", " ")

    def __repr__(self) -> str:
        return str(self)

    def get_value(
        self, endianness: str, data: Streamable
    ) -> Optional[BaseType.value_type]:
        # TODO: check for endian ability before creating fmt_string?
        fmt_string = f"{endianness}{self.fmt}"

        (value,) = struct.unpack(fmt_string, data.read(self.size))

        if value == self.invalid_value:
            return None

        return value


BASE_TYPES = {
    0x00: BaseType(
        number=0,
        endian_ability=0,
        name="enum",
        invalid_value=0xFF,
        size=1,
        fmt="B",
        value_type=int,
        comment="",
    ),
    0x01: BaseType(
        number=1,
        endian_ability=0,
        name="sint8",
        invalid_value=0x7F,
        size=1,
        fmt="b",
        value_type=int,
        comment="2’s complement format",
    ),
    0x02: BaseType(
        number=2,
        endian_ability=0,
        name="uint8",
        invalid_value=0xFF,
        size=1,
        fmt="B",
        value_type=int,
        comment="2’s complement format",
    ),
    0x83: BaseType(
        number=3,
        endian_ability=1,
        name="sint16",
        invalid_value=0x7FFF,
        size=2,
        fmt="h",
        value_type=int,
        comment="2’s complement format",
    ),
    0x84: BaseType(
        number=4,
        endian_ability=1,
        name="uint16",
        invalid_value=0xFFFF,
        size=2,
        fmt="H",
        value_type=int,
        comment="",
    ),
    0x85: BaseType(
        number=5,
        endian_ability=1,
        name="sint32",
        invalid_value=0x7FFFFFFF,
        size=4,
        fmt="i",
        value_type=int,
        comment="2’s complement format",
    ),
    0x86: BaseType(
        number=6,
        endian_ability=1,
        name="uint32",
        invalid_value=0xFFFFFFFF,
        size=4,
        fmt="I",
        value_type=int,
        comment="",
    ),
    0x07: BaseType(
        number=7,
        endian_ability=0,
        name="string",
        invalid_value=0x00,
        size=1,
        fmt="s",
        value_type=str,
        comment="Null terminated string encoded in UTF-8 format",
    ),
    0x88: BaseType(
        number=8,
        endian_ability=1,
        name="float32",
        invalid_value=0xFFFFFFFF,
        size=4,
        fmt="f",
        value_type=float,
        comment="",
    ),
    0x89: BaseType(
        number=9,
        endian_ability=1,
        name="float64",
        invalid_value=0xFFFFFFFFFFFFFFFF,
        size=8,
        fmt="d",
        value_type=float,
        comment="",
    ),
    0x0A: BaseType(
        number=10,
        endian_ability=0,
        name="uint8z",
        invalid_value=0x00,
        size=1,
        fmt="B",
        value_type=int,
        comment="",
    ),
    0x8B: BaseType(
        number=11,
        endian_ability=1,
        name="uint16z",
        invalid_value=0x0000,
        size=2,
        fmt="H",
        value_type=int,
        comment="",
    ),
    0x8C: BaseType(
        number=12,
        endian_ability=1,
        name="uint32z",
        invalid_value=0x00000000,
        size=4,
        fmt="I",
        value_type=int,
        comment="",
    ),
    0x0D: BaseType(
        number=13,
        endian_ability=1,
        name="byte",
        invalid_value=0xFF,
        size=1,
        fmt="B",
        value_type=int,
        comment="Array of bytes. Field is invalid if all bytes are invalid.",
    ),
    0x8E: BaseType(
        number=14,
        endian_ability=1,
        name="sint64",
        invalid_value=0x7FFFFFFFFFFFFFFF,
        size=8,
        fmt="q",
        value_type=int,
        comment="2’s complement format",
    ),
    0x8F: BaseType(
        number=15,
        endian_ability=1,
        name="uint64",
        invalid_value=0xFFFFFFFFFFFFFFFF,
        size=8,
        fmt="Q",
        value_type=int,
        comment="",
    ),
    0x90: BaseType(
        number=16,
        endian_ability=1,
        name="uint64z",
        invalid_value=0x0000000000000000,
        size=8,
        fmt="Q",
        value_type=int,
        comment="",
    ),
}



from __future__ import annotations

from datetime import datetime, timezone

FIT_EPOCH = 631065600


def rollover_timestamp(previous_timestamp: int, offset: int) -> int:
    """
    Apply the compressed timestamp offset to the previous timestamp

    The offset is a 5-bit offset which rolls over every 32 seconds
    (0b11111 == 31). This means that consecutive compressed timestamps may
    never be more than 32 seconds apart

    The actual timestamp is calculated by concatenating the most significant 27 bits
    of the previous timestamp value and the 5 bit value of the offset field
    """
    max_length = 0b1111  # 0x0000001F

    if offset >= previous_timestamp & max_length:
        # Offset value is greater than least significant 5 bits of previous timestamp
        timestamp = (previous_timestamp & 0xFFFFFFE0) + offset
    else:
        # Offset is less than least significant 5 bits of previous timestamp
        timestamp = (previous_timestamp & 0xFFFFFFE0) + offset + 0x20

    return timestamp


def datetime_from_timestamp(timestamp: int) -> datetime:
    """
    Create a datetime from a timestamp using the Garmin FIT epoch, in UTC.
    """
    dt = datetime.utcfromtimestamp(timestamp + FIT_EPOCH)
    dt = dt.replace(tzinfo=timezone.utc)
    return dt



import struct
from io import BytesIO

from fittie.utils.datastream import Streamable
from fittie.utils.exceptions import DecodeException
from fittie.fitfile.crc import calculate_crc

DEFAULT_CRC = 0x0000


class Header:
    """
    File header which provides data about the FIT File

    Minimum size is 12 bytes, but a 14 byte header is preferred.

    Computing the CRC is optional and 0x0000 is a permissible CRC value
    """

    fmt: str = "BBHI4s"
    length: int
    protocol_version: int
    profile_version: int
    data_size: int
    data_type: str
    crc: int

    def __init__(
        self,
        length: int,
        protocol_version: int,
        profile_version: int,
        data_size: int,
        data_type: str,
        crc: int,
    ) -> None:
        self.length = length
        self.protocol_version = protocol_version
        self.profile_version = profile_version
        self.data_size = data_size
        self.data_type = data_type
        self.crc = crc

    def encode(self) -> bytes:
        """Encode the header into bytes"""
        values = (
            self.length,
            self.protocol_version,
            self.profile_version,
            self.data_size,
            self.data_type.encode("utf-8"),
        )

        if self.length == 14:
            self.fmt += "H"  # add additional 2 bytes for CRC
            values += (self.crc,)

        return struct.pack(self.fmt, *values)

    def __str__(self) -> str:
        return (
            f"Header:{self.length=}{self.protocol_version=}"
            f"{self.profile_version=}{self.data_size=}"
            f"{self.data_type=}{self.crc=}"
        ).replace("self.", " ")


def decode_header(data: Streamable) -> Header:
    """
    Reads a FIT file header from the provided data
    """
    try:
        # Read first 12 bytes to determine header size
        header_data = data.read(12)
        header_reader = BytesIO(header_data)

        (length,) = struct.unpack("B", header_reader.read(1))
        (protocol_version,) = struct.unpack("B", header_reader.read(1))
        (profile_version,) = struct.unpack("H", header_reader.read(2))
        (data_size,) = struct.unpack("I", header_reader.read(4))
        data_type = b"".join(struct.unpack("4s", header_reader.read(4))).decode("utf-8")

        if length == 14:
            (crc,) = struct.unpack("H", data.read(2))
            calculated_crc = calculate_crc(header_data)

            if crc != calculated_crc:
                raise DecodeException(
                    detail="invalid crc checksum in file header", position=0
                )
        else:
            crc = DEFAULT_CRC
    except struct.error as exc:
        raise DecodeException(
            detail="could not decode header with provided data",
            position=data.tell(),
        ) from exc

    header = Header(
        length=length,
        protocol_version=protocol_version,
        profile_version=profile_version,
        data_size=data_size,
        data_type=data_type,
        crc=crc,
    )

    # TODO: compute CRC and check

    return header



from unittest import mock
from functools import cached_property

from fittie.utils.iterable import IterableMixin


class Foo(IterableMixin):
    lst: list

    def __init__(self, lst):
        self.lst = lst

    @property
    def _iter_collection(self):
        return self.lst


def compute() -> list:
    return [1, 2, 3, 4, 5]


class FooCached(IterableMixin):
    @cached_property
    def _iter_collection(self):
        return compute()


def test_iterable_non_cached_property():
    lst = [1, 2, 3, 4, 5]
    foo = Foo(lst)

    for idx, value in enumerate(foo):
        assert value == lst[idx]


def test_iterable_cached_property():
    foo = FooCached()
    lst = [1, 2, 3, 4, 5]

    with mock.patch(
        "tests.fittie.utils.test_iterable.compute", return_value=lst
    ) as mock_compute:
        for idx, value in enumerate(foo):
            assert value == lst[idx]

        mock_compute.assert_called_once()

    for idx, value in enumerate(foo):
        assert value == lst[idx]



from __future__ import annotations  # Added for type hints

import functools
import itertools
import logging
import os
import struct

from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Optional, Union, Iterable

from fittie.utils.datastream import DataStream, Streamable
from fittie.utils.exceptions import DecodeException
from fittie.fitfile.data_message import DataMessage
from fittie.fitfile.definition_message import DefinitionMessage
from fittie.fitfile.field_description import FieldDescription
from fittie.fitfile.header import Header, decode_header
from fittie.fitfile.profile.fit_types import FIT_TYPES
from fittie.fitfile.profile.mesg_nums import MESG_NUMS
from fittie.fitfile.records import read_record_header, read_record
from fittie.fitfile.util import datetime_from_timestamp
from fittie.utils.iterable import IterableMixin

LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()
logging.basicConfig(level=LOGLEVEL)
logger = logging.getLogger("fittie")


class FitFile(IterableMixin):

    header: Header
    data_messages: dict[str, list[DataMessage]]
    local_message_definitions: dict[int, DefinitionMessage] = {}
    developer_data: dict[int, dict[str, Any]] = {}  # TODO: add typing

    def __init__(
        self,
        header: Header,
        data_messages: dict[str, list[DataMessage]],
        local_message_definitions: dict[int, DefinitionMessage],
        developer_data: dict[int, dict[str, Any]],
    ):
        self.header = header
        self.data_messages = data_messages
        self.local_message_definitions = local_message_definitions
        self.developer_data = developer_data

    @functools.cached_property
    def _iter_collection(self) -> Iterable:
        return list(itertools.chain(*self.data_messages.values()))

    @property
    def average_heart_rate(self) -> Optional[int]:
        """Get average heart rate from data messages, if any"""
        heart_rates: list[int] = []

        if not (records := self.data_messages.get("record")):
            return None

        for message in records:
            if (heart_rate := message.fields.get("heart_rate")) is not None:
                heart_rates.append(heart_rate)
        if not heart_rates:
            return None

        return int(sum(heart_rates) / len(heart_rates))

    @property
    def file_id(self) -> Optional[dict[str, Any]]:
        """
        Get file id information.

        Raw values from the FIT file will be filled with information from the Garmin
        FIT SDK Fit Types
        """
        if not (file_id_messages := self.data_messages.get("file_id")):
            raise ValueError(
                "no file_id message detected, FIT file is possible incorrect"
            )

        file_id = {}

        # Should be just one file_id, but to be sure use latest from list
        for key, value in file_id_messages[-1].fields.items():
            if key == "time_created":
                file_id[key] = datetime_from_timestamp(value)
            elif key == "type":
                file_id[key] = FIT_TYPES["file"]["values"][value]["value_name"]
            elif key == "manufacturer":
                file_id[key] = FIT_TYPES["manufacturer"]["values"][value]["value_name"]
            else:
                file_id[key] = value

        return file_id

    @property
    def file_type(self) -> str:
        """
        Returns the file type of the FIT File. The file type is retrieved from the
        file_id message, which is always present in a FIT file.
        """
        return self.file_id["type"]

    @property
    def available_message_types(self) -> list[str]:
        """Returns a list of all message types that this FIT file contains"""
        return list(self.data_messages.keys())

    def get_messages_by_type(self, message_type: str) -> list[DataMessage]:
        """
        Returns all messages of the provided type, if the FIT file contains these
        messages. If not, it will return an empty list.

        If the provided message type is unknown, a ValueError will be raised.
        """
        if message_type not in MESG_NUMS.values():
            raise ValueError(f"unknown message type '{message_type}' received")

        return self.data_messages.get(message_type, [])


def decode(
    source: Union[str, Path, Streamable], calculate_crc: Optional[bool] = True
) -> FitFile:
    """
    Decode a fit file
    """
    with DataStream(source) as data:
        if not calculate_crc:
            # Don't calculate checksum
            data.should_calculate_crc = False

        header = decode_header(data)

        logger.debug(header)

        local_message_definitions: dict[int, DefinitionMessage] = {}
        developer_data: dict[int, dict[str, Any]] = {}
        messages: DefaultDict[str, list[DataMessage]] = defaultdict(list)

        while data.tell() < header.length + header.data_size:
            # Read record header
            record_header = read_record_header(data)

            # Read message
            message = read_record(
                record_header,
                local_message_definitions.get(record_header.local_message_type),
                developer_data,
                data,
            )
            logger.debug(message)

            # Assign message to correct collection
            if record_header.is_compressed_timestamp_message:
                # TODO:
                ...
            elif record_header.is_developer_data or record_header.is_definition_message:
                # TODO: check if this can be merged with is_definition_message
                local_message_definitions[record_header.local_message_type] = message
            else:
                if (
                    global_message_type := local_message_definitions.get(
                        record_header.local_message_type
                    ).global_message_type
                ) == 207:
                    # Add developer data index
                    index = message.fields["developer_data_index"]
                    developer_data[index] = message.fields
                    developer_data[index].update({"fields": {}})
                elif global_message_type == 206:
                    # Add field descriptions
                    index = message.fields["developer_data_index"]
                    field = FieldDescription(**message.fields)
                    developer_data[index]["fields"][
                        field.field_definition_number
                    ] = field

                messages[MESG_NUMS[global_message_type]].append(message)

        calculated_crc = data.calculated_crc
        (crc,) = struct.unpack("H", data.read(2))

        if calculate_crc and crc != calculated_crc:
            raise DecodeException(
                detail=(
                    "the calculated crc does not match the crc at the end of the file"
                ),
                position=data.tell(),
            )

    fitfile = FitFile(
        header=header,
        data_messages=messages,
        local_message_definitions=local_message_definitions,
        developer_data=developer_data,
    )

    return fitfile



from io import BytesIO

from fittie.fitfile.data_message import decode_data_message
from fittie.fitfile.records import RecordHeader


def test_decode_data_message(record_1_definition_message):
    # Expect record 2
    header = RecordHeader(
        is_definition_message=False,
        is_developer_data=False,
        is_compressed_timestamp_message=False,
        local_message_type=0
    )
    data = b'\x04\x0f\x00\x16\x00\xd2\x04\x00\x00(\xc6\n%'
    data_message = decode_data_message(
        header,
        record_1_definition_message,
        developer_data={},
        data=BytesIO(data)
    )
    assert data_message.fields == {
        "type": 4,
        "manufacturer": 15,
        "product": 22,
        "serial_number": 1234,
        "time_created": 621463080,
    }


def test_decode_data_message_with_developer_fields(record_5_definition_message):
    # Expect record 6
    header = RecordHeader(
        is_definition_message=False,
        is_developer_data=False,
        is_compressed_timestamp_message=False,
        local_message_type=0
    )
    data = (
        b'\x00\x00\x01'
        b'doughnuts earned\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'doughnuts\x00\x00\x00\x00\x00\x00\x00'
    )
    data_message = decode_data_message(
        header,
        record_5_definition_message,
        developer_data={},
        data=BytesIO(data)
    )
    assert data_message.fields == {
        "developer_data_index": 0,
        "fit_base_type_id": 1,
        "field_definition_number": 0,
        "field_name": "doughnuts earned",
        "units": "doughnuts"
    }
